// This unit is written and copyright (c) 1997 Kirill Shpitsa
// See langsupp.hlp for more information.
// Please redistribute original .zip file
// email: <kirill@shkirill.kiev.ua>

{$IFDEF VER100}
{$DEFINE D3}
{$ENDIF}

unit
   langsupp;

interface

uses
   Windows,
   Messages,
   SysUtils,
   Classes,
   Graphics,
   Controls,
   Forms,
   Dialogs,
   DsgnIntf,
   TypInfo;

type
   TLanguage=type string;
   TLangSupport=
      class(TComponent)
      private
         FActiveLanguage:TLanguage;
         FLanguages:TStrings;
         LangStrings:TStringList;
         FSaveFont:boolean;

         procedure ReadData(Reader:TReader);
         procedure SetActiveLanguage(Value:TLanguage);
         procedure SetLanguages(Value:TStrings);
         procedure WriteData(Writer:TWriter);

         procedure LanguagesChange(Sender:TObject);
      protected
         procedure DefineProperties(Filer:TFiler);override;
         procedure Loaded;override;
         procedure Notification
            (AComponent:TComponent;Operation:TOperation);override;

         procedure LoadLanguage(id:string);
         procedure SaveLanguage(id:string);
      public
         constructor Create(AOwner:TComponent);override;
         destructor Destroy;override;
      published
         property ActiveLanguage:TLanguage
            read FActiveLanguage write SetActiveLanguage;
         property Languages:TStrings read FLanguages write SetLanguages
            stored false;
         property SaveFont:boolean read FSaveFont write FSaveFont;
      end;

procedure SetActiveLanguage(Value:string);
procedure Register;

implementation

var
   InstanceList:TList;

{---miscellaneous--------------------------------------------------------}

function GetComponentFont(AComponent:TComponent):string;
var
   p:PPropInfo;
begin
   if GetPropInfo(AComponent.ClassInfo,'caption')<>nil then
   begin
      p:=GetPropInfo(AComponent.ClassInfo,'font');
      if Assigned(p) and
         (p^.PropType{$IFDEF D3}^{$ENDIF}.Kind=tkClass) and
         (GetTypeData(p^.PropType{$IFDEF D3}^{$ENDIF}).ClassType.InheritsFrom(TFont))
         then
            Result:=TFont(pointer(GetOrdProp(AComponent,p))).Name
         else
            Result:='';
   end else Result:='';
end;{GetComponentFont}

procedure SetComponentFont(AComponent:TComponent;font:string);
var
   p:PPropInfo;
begin
   if GetPropInfo(AComponent.ClassInfo,'caption')<>nil then
   begin
      p:=GetPropInfo(AComponent.ClassInfo,'font');
      if Assigned(p) and
         (p^.PropType{$IFDEF D3}^{$ENDIF}.Kind=tkClass) and
         (GetTypeData(p^.PropType{$IFDEF D3}^{$ENDIF}).ClassType.InheritsFrom(TFont)) then
         TFont(pointer(GetOrdProp(AComponent,p))).Name:=font;
   end;
end;{SetComponentFont}

function GetComponentProp(AComponent:TComponent;name:string):string;
var
   p:PPropInfo;
begin
   p:=GetPropInfo(AComponent.ClassInfo,name);
   if Assigned(p)
      then Result:=GetStrProp(AComponent,p)
      else Result:='';
end;{GetComponentProp}

procedure SetComponentProp(AComponent:TComponent;const Name,Value:string);
var
   p:PPropInfo;
begin
   if Value='' then exit;
   p:=GetPropInfo(AComponent.ClassInfo,name);
   if Assigned(p) then SetStrProp(AComponent,p,Value);
end;{SetComponentProp}

function CaptionedComponent(AComponent:TComponent):boolean;
begin
   Result:=(AComponent.Name<>'') and
      ((GetPropInfo(AComponent.ClassInfo,'caption')<>nil) or
       (GetPropInfo(AComponent.ClassInfo,'hint')<>nil));
end;{CaptionedComponent}

{---TLangString&TLangList------------------------------------------------}

type
   TLangString=
      class
         Component:TComponent;
         ComponentName:string;
         Caption,
         Hint,
         Font:string;
      end;

   TLangList=
      class(TList)
         FOwner:TComponent;
         constructor Create(AOwner:TComponent);
         destructor Destroy;override;
         procedure AddData(Component,Caption,hint,Font:string);
         procedure RemoveData(Component:TComponent);
         procedure ClearAll;
         procedure Link;
      end;

constructor TLangList.Create(AOwner:TComponent);
begin
   inherited Create;
   FOwner:=AOwner;
end;{TLangList.Create}

destructor TLangList.Destroy;
begin
   ClearAll;
   inherited Destroy;
end;{TLangList.Destroy}

procedure TLangList.AddData(Component,Caption,Hint,Font:string);
var
   t:TLangString;
begin
   t:=TLangString.Create;
   t.ComponentName:=Component;
   t.Caption:=Caption;
   t.Hint:=Hint;
   t.Font:=Font;
   Add(t);
end;{TLangList.AddData}

procedure TLangList.ClearAll;
var
   i:integer;
begin
   for i:=0 to Count-1 do
      TObject(Items[i]).Free;

   inherited Clear;
end;{TLangList.ClearAll}

procedure TLangList.Link;
var
   i:integer;
begin
   for i:=0 to Count-1 do
      with TLangString(Items[i]) do
      begin
         if ComponentName='.'
            then Component:=FOwner
            else Component:=FOwner.FindComponent(ComponentName);
         if not Assigned(Component) then
         begin
            Free;
            Delete(i);
         end;
         ComponentName:='';
      end;
   Pack;
end;{TLangList.Link}

procedure TLangList.RemoveData(Component:TComponent);
var
   i:integer;
begin
   i:=0;
   while (i<Count) and
      (TLangString(Items[i]).Component<>Component) do inc(i);

   if i<Count then
      with TLangString(Items[i]) do
      begin
         Free;
         Remove(Items[i]);
      end;
end;{TLangList.RemoveData}

{---TLanguageSelector----------------------------------------------------}

type
   TLanguageSelector=
      class(TPropertyEditor)
         function GetAttributes:TPropertyAttributes;override;
         function GetValue:string;override;
         procedure GetValues(Proc:TGetStrProc);override;
         procedure SetValue(const Value:string);override;
      end;

function TLanguageSelector.GetAttributes:TPropertyAttributes;
begin
   Result:=[paMultiSelect,paValueList];
end;{TLanguageSelector.GetAttributes}

function TLanguageSelector.GetValue:string;
begin
   Result:=GetStrValue;
end;{TLanguageSelector.GetValue}

procedure TLanguageSelector.GetValues(Proc:TGetStrProc);
var
   i:integer;
   t:TLangSupport;
begin
   t:=GetComponent(0) as TLangSupport;
   for i:=0 to t.FLanguages.Count-1 do
      Proc(t.FLanguages[i]);
end;{TLanguageSelector.GetValues}

procedure TLanguageSelector.SetValue(const Value:string);
begin
   SetStrValue(Value);
end;{TLanguageSelector.SetValue}

{---TLangSupport---------------------------------------------------------}

constructor TLangSupport.Create(AOwner:TComponent);
var
   i:integer;
   SecondInstance:boolean;
begin
   SecondInstance:=false;
   with AOwner do
      for i:=0 to ComponentCount-1 do
         if Components[i] is TLangSupport then
         begin
            SecondInstance:=true;
            break;
         end;

   inherited Create(AOwner);

   if not(AOwner is TForm) then
      raise Exception.Create(
         ClassName+' can be placed on TForm only');

   if SecondInstance then
      raise Exception.Create(
         'Only single instance of '+ClassName+
         ' is permitted to exist on form');

   FLanguages:=TStringList.Create;
   LangStrings:=TStringList.Create;
   with TStringList(FLanguages) do
   begin
      Sorted:=true;
      Duplicates:=dupIgnore;
      if csDesigning in ComponentState then
         OnChange:=LanguagesChange;
   end;

   InstanceList.Add(Self);
end;{TLangSupport.Create}

destructor TLangSupport.Destroy;
var
   i:integer;
begin
   if Assigned(LangStrings) then
      for i:=0 to LangStrings.Count-1 do
         TLangList(LangStrings.Objects[i]).Free;
   LangStrings.Free;
   FLanguages.Free;

   if Assigned(InstanceList) then
      InstanceList.Remove(Self);
   inherited Destroy;
end;{TLangSupport.Destroy}

procedure TLangSupport.DefineProperties(Filer:TFiler);
begin
   inherited DefineProperties(Filer);

   Filer.DefineProperty('Data',ReadData,WriteData,true);
end;{TLangSupport.DefineProperties}

procedure TLangSupport.Loaded;
var
   i:integer;
begin
   inherited Loaded;

   for i:=0 to LangStrings.Count-1 do
      TLangList(LangStrings.Objects[i]).Link;
end;{TLangSupport.Loaded}

procedure TLangSupport.Notification(AComponent:TComponent;Operation:TOperation);
var
   i:integer;
begin
   inherited Notification(AComponent,Operation);

   if (csDesigning in ComponentState) and
      (Operation=opRemove) and
      (AComponent<>Self) then
   begin
      with LangStrings do
         for i:=0 to Count-1 do
            TLangList(Objects[i]).RemoveData(AComponent);
   end;
end;{TLangSupport.Notification}

procedure TLangSupport.LoadLanguage(id:string);
var
   i,j:integer;
begin
   i:=LangStrings.IndexOf(id);
   if i=-1 then exit;
   with LangStrings.Objects[i] as TLangList do
      for j:=0 to Count-1 do
         with TLangString(Items[j]) do
         begin
            SetComponentProp(Component,'Caption',Caption);
            SetComponentProp(Component,'Hint',Hint);
            if FSaveFont then
               SetComponentFont(Component,Font);
         end;
end;{TLangSupport.LoadLanguage}

procedure TLangSupport.ReadData(Reader:TReader);
var
   LangId,s1,s2,s3,s4:string;
   list:TLangList;
begin
   with Reader do
   begin
      ReadListBegin;
      while not EndOfList do
      begin
         LangId:=ReadString;
         FLanguages.Add(LangId);
         list:=TLangList.Create(Owner);
         LangStrings.AddObject(LangId,list);

         ReadListBegin;
         while not EndOfList do
         begin
            ReadListBegin;
            s1:=ReadString;
            s2:=ReadString;
            s3:=ReadString;
            if not EndOfList
               then s4:=ReadString else s4:='';
            ReadListEnd;

            list.AddData(s1,s2,s3,s4);
         end;
         ReadListEnd;
      end;
      ReadListEnd;
   end;
end;{TLangSupport.ReadData}

procedure TLangSupport.SaveLanguage(id:string);
var
   i:integer;
   list:TLangList;
begin
   i:=LangStrings.IndexOf(id);
   if i>=0
      then TLangList(LangStrings.Objects[i]).ClearAll
      else i:=LangStrings.AddObject(id,TLangList.Create(Owner));
   list:=LangStrings.Objects[i] as TLangList;

   if CaptionedComponent(Owner) then
      list.AddData('.',
         GetComponentProp(Owner,'Caption'),
         GetComponentProp(Owner,'Hint'),
         GetComponentFont(Owner));

   with Owner do
      for i:=0 to ComponentCount-1 do
         if CaptionedComponent(Components[i]) then
            list.AddData(Components[i].Name,
               GetComponentProp(Components[i],'Caption'),
               GetComponentProp(Components[i],'Hint'),
               GetComponentFont(Components[i]));
   list.link;
end;{TLangSupport.SaveLanguage}

procedure TLangSupport.SetActiveLanguage(Value:TLanguage);
begin
   if FActiveLanguage<>'' then SaveLanguage(FActiveLanguage);
   FActiveLanguage:=Value;
   LoadLanguage(Value);
end;{TLangSupport.SetActiveLanguage}

procedure TLangSupport.SetLanguages(Value:TStrings);
begin
   FLanguages.Assign(Value);
end;{TLangSupport.SetLanguages}

procedure TLangSupport.WriteData(Writer:TWriter);
var
   i,j:integer;
begin
   SaveLanguage(ActiveLanguage);
   with Writer do
   begin
      WriteListBegin;
      for i:=0 to FLanguages.Count-1 do
      begin
         WriteString(FLanguages[i]);
         WriteListBegin;
         if (Languages[i]<>ActiveLanguage) and
            (LangStrings.IndexOf(Languages[i])<>-1) then
            with TLangList(LangStrings.Objects[i]) do
               for j:=0 to Count-1 do
                  with TLangString(Items[j]) do
                  begin
                     WriteListBegin;
                     if Component=Owner
                        then WriteString('.')
                        else WriteString(Component.Name);
                     WriteString(Caption);
                     WriteString(Hint);
                     if FSaveFont then
                        WriteString(Font);
                     WriteListEnd;
                  end;
         WriteListEnd;
      end;
      WriteListEnd;
   end;
end;{TLangSupport.WriteData}

procedure TLangSupport.LanguagesChange(Sender:TObject);
var
   i:integer;
begin
   i:=0;
   with FLanguages do
      while i<Count do
         if Strings[i]=''
            then Delete(i)
            else inc(i);
   i:=0;
   while i<LangStrings.Count do
      if FLanguages.IndexOf(LangStrings[i])=-1 then
      begin
         TLangList(LangStrings.Objects[i]).Free;
         LangStrings.Delete(i);
      end else inc(i);
end;{TLangSupport.LanguagesChange}

procedure SetActiveLanguage(Value:string);
var
   i:integer;
begin
   with InstanceList do
      for i:=0 to Count-1 do
         TLangSupport(Items[i]).ActiveLanguage:=Value;
end;{SetActiveLanguage}

procedure Register;
begin
   RegisterComponents('Samples', [TLangSupport]);

   RegisterPropertyEditor(TypeInfo(TLanguage),TLangSupport,'',
      TLanguageSelector);
end;{Register}

initialization
   InstanceList:=TList.Create;
finalization
   InstanceList.Free;
end.
